#    Copyright (C) 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import sys

if __name__ == "__main__":

    sys.path.insert(0, ".")

    # from demo.demo_mysql import demo
    # from demo.demo_postgresql import demo
    # from demo.demo_sqlite import demo
    # from temp.try_pg_01 import demo
    from tests.test_db import test_mysql, test_posrtgresql, test_sqlite

    test_mysql()
    test_sqlite()
    test_posrtgresql()

import re

import sqlite3 as db_sqlite_connector
import mysql.connector as db_mysql_connector
import psycopg2 as db_postgresql_connector


from zzdb.utils import int_, is_sub_list
from zzdb.cursor import ZzMysqlCursor, ZzSqliteCursor, ZzPostgresqlCursor


class ZzDb:
    def __init__(
        self,
        db_engine_name,
        user=None,
        password=None,
        host=None,
        database_name=None,
        port=None,
        guest_mode=False,
    ):
        self.db_engine_name = db_engine_name
        self.user = user
        self.password = password
        self.host = host
        self.database_name = database_name
        self.port = port
        self.guest_mode = guest_mode
        self.connection = None

        if self.db_engine_name not in ["mysql", "sqlite3", "postgresql"]:
            raise Exception(f"Sorry, wrong DBAPI engine - {self.db_engine_name}")

        self.last_sql_error = ""
        self.last_sql = ""
        self.last_error_data = {}
        self.last_record = ""
        self.migrate_error_list = []

        self.db_schema = None
        self.connection = None

        if self.db_engine_name == "mysql":
            self.db_api_engine = db_mysql_connector
            self.db_cursor_class = ZzMysqlCursor
        elif self.db_engine_name == "postgresql":
            self.db_api_engine = db_postgresql_connector
            self.db_cursor_class = ZzPostgresqlCursor
        elif self.db_engine_name == "sqlite3":
            self.db_api_engine = db_sqlite_connector
            self.db_cursor_class = ZzSqliteCursor

        try:
            self._connect()
            return
        except Exception:
            pass  # Do nothing - give chance to screate database!
        self.create()
        self._connect()

    def _connect(self):
        self.connection = self.connect(
            user=self.user,
            password=self.password,
            host=self.host,
            database_name=self.database_name,
            port=self.port,
        )

    def create(self):
        """
        Take chance to create database
        """
        admin_database_name = {"mysql": "mysql", "postgresql": "postgres"}[self.db_engine_name]
        try:
            self.connection = self.connect(
                user=self.user,
                password=self.password,
                host=self.host,
                database_name=admin_database_name,
                port=self.port,
            )
        except Exception:
            print(sys.exc_info())
        if self.db_engine_name == "mysql":
            self._cursor(sql=f"CREATE DATABASE IF NOT EXISTS {self.database_name}")
            self._cursor(sql=f"GRANT ALL PRIVILEGES ON {self.database_name}.* TO '{self.user}'")
        elif self.db_engine_name == "postgresql":
            self._cursor(sql=f"CREATE DATABASE {self.database_name} WITH OWNER = {self.user}")
            self._cursor(sql=f"GRANT ALL PRIVILEGES ON DATABASE {self.database_name}.* TO {self.user}")
        self.connection.close()
        return True

    def connect(
        self,
        user=None,
        password=None,
        host=None,
        database_name=None,
        port=None,
    ):
        connection = None
        if self.db_engine_name == "mysql":
            connection = self.db_api_engine.connect(
                user=user,
                password=password,
                host=host if host else "localhost",
                port=port if port else 3306,
                database=database_name,
            )
        elif self.db_engine_name == "postgresql":
            connection = self.db_api_engine.connect(
                user=user,
                password=password,
                host=host if host else "localhost",
                port=port if port else 5432,
                database=database_name,
            )
            connection.autocommit = True
        elif self.db_engine_name == "sqlite3":
            connection = self.db_api_engine.connect(database=self.database_name, isolation_level=None)
        return connection

    def close(self):
        self.connection.close()

    def get_primary_key_columns(self, table_name=""):
        """returns database primary columns for given table"""
        if self.db_engine_name in ("mysql", "postgresql"):
            return self.get_database_columns(table_name, "column_key='PRI'")
        elif self.db_engine_name == "sqlite3":
            return self.get_database_columns(table_name, "pk=1")

    def get_database_columns(self, table_name="", filter=""):
        """returns database columns for given table"""
        cols = {}
        sql = self.db_cursor_class.get_table_columns_sql(table_name, filter, self.database_name)
        for x in self.cursor(sql).records():
            if "name" in x:
                if "datalen" not in x:  # SQLITE
                    if "(" in x["datatype"]:
                        ln = re.search(r"\((.*?)\)", x["datatype"]).group(1)
                        if "," not in ln:
                            x["datalen"] = ln
                            x["datadec"] = "0"
                        else:
                            x["datalen"] = ln.split(",")[0]
                            x["datadec"] = ln.split(",")[1]
                        x["datatype"] = x["datatype"].split("(")[0]
                    else:
                        x["datalen"] = "0"
                        x["datadec"] = "0"
                cols[x["name"]] = x
        return cols

    def get_tables(self, table_name=""):
        """Returns a list of tables names from database"""
        table_select_clause = f" and TABLE_NAME='{table_name}'" if table_name else ""
        sql = self.db_cursor_class.get_table_names_sql(table_select_clause, self.database_name)
        return [x["table_name"] for x in self.cursor(sql).records()]

    def set_schema(self, db_schema):
        """assign and migrate schema to database"""
        self.db_schema = db_schema
        return self.migrate_schema()

    def migrate_schema(self):
        """creates (no alter) tables and columns in a physical database"""
        if self.db_schema is None:
            return
        self.migrate_error_list = []
        for table in self.db_schema.get_schema_table_attr():
            # column that are already in
            database_columns = self.get_database_columns(table)
            # print(self.db_schema.get_schema_table_attr(table))
            schema_columns = self.db_schema.get_schema_table_attr(table)
            if not self.guest_mode:
                self._add_zz_columns(schema_columns)
            for column in schema_columns:
                # self.db_schema.refresh_foreignkey_column(table, column)
                colDic = self.db_schema.get_schema_table_attr(table, column)
                if column not in database_columns:  # new column
                    colDic["table"] = table
                    colDic["column"] = column
                    self.create_column(colDic)
                else:  # change schema as it is in database
                    colDic["datalen"] = database_columns[column]["datalen"]
                    colDic["datadec"] = database_columns[column]["datadec"]
        return True

    def _add_zz_columns(self, schema_columns):
        """adding zz-columns to the zz-schema"""
        schema_columns["zz_time"] = {"datatype": "bigint"}
        schema_columns["zz_mode"] = {"datatype": "char", "datalen": 1}
        # schema_columns["update_time"] = {"datatype": "bigint"}
        # schema_columns["insert_session_id"] = {"datatype": "int"}
        # schema_columns["update_session_id"] = {"datatype": "int"}
        # schema_columns["user_lock"] = {"datatype": "char", "datalen": 1}

    def create_column(self, column_definition):
        """migrate given 'column_definition' to database"""

        column_definition["datadec"] = column_definition.get("datadec", 0)
        column_definition["primarykey"] = "PRIMARY KEY" if column_definition.get("pk", "") else ""
        if column_definition.get("to_table") and column_definition.get("to_column"):
            # pull attributes from primary table
            primary_column_definition = self.db_schema.get_schema_table_attr(
                column_definition["to_table"], column_definition["to_column"]
            )
            for x in ["datatype", "datalen", "datadec"]:
                column_definition[x] = primary_column_definition[x]

        if "datatype" not in column_definition:
            return None
        datatype = column_definition["datatype"].upper()
        if datatype[:3] in ["NUM", "DEC"]:
            column_definition["datatype"] = "NUMERIC"
            column_definition["default"] = "DEFAULT '0'"
            column_definition["size"] = "({datalen},{datadec})".format(**column_definition)
        elif "INT" in datatype:
            column_definition["default"] = "DEFAULT '0'"
            column_definition["size"] = ""
        elif "CHAR" in datatype:
            column_definition["default"] = "DEFAULT ''"
            column_definition["size"] = (
                "({datalen})".format(**column_definition) if column_definition["datalen"] else ""
            )
        elif "DATE" in datatype:
            column_definition["default"] = "DEFAULT '0000-00-00'"
            column_definition["size"] = ""
        elif "TEXT" in datatype:
            column_definition["default"] = ""
            column_definition["size"] = ""
        else:
            column_definition["size"] = ""
            column_definition["default"] = ""

        sql_column_text = """{column} {datatype} {size} {primarykey} {default}""".format(**column_definition)
        table = column_definition["table"]
        if table in self.get_tables():
            sqlCmd = f"ALTER TABLE {table} ADD {sql_column_text}"
        else:
            sqlCmd = f"CREATE TABLE {table} ({sql_column_text})"

        self._cursor(sqlCmd)
        if self.last_sql_error != "":
            self.migrate_error_list.append(self.last_sql_error)
            return False
        # add the same colunm into log_ table
        if not self.guest_mode and not table.upper().startswith("LOG_"):
            log_column_definition = dict(column_definition)
            log_column_definition["pk"] = None
            log_column_definition["ai"] = None
            log_column_definition["uk"] = None
            log_column_definition["table"] = "log_" + log_column_definition["table"]
            self.create_column(log_column_definition)
        return True

    def _sqlite_patch(self, sql, record, table_columns):
        """Adapt sql statement for sqlite - convert str to int, replace placeholder character to ?"""
        for x in record:
            if "int" in table_columns.get(x, {}).get("type", ""):
                try:
                    record[x] = int_(record[x])
                except Exception:
                    print(table_columns.get(x, {}).get("type", ""), record[x])
        sql = sql.replace("%s", "?")
        return sql

    def _check_record_for_numbers(self, table, record):
        """ "makes sure that all number columns value is not blank string"""
        for x in record:
            if record[x] != "":
                continue
            datatype = self.db_schema.get_schema_table_attr(table, x).get("datatype")
            if "int" in datatype or "dec" in datatype or "num" in datatype:
                record[x] = "0"

    def insert(self, table_name="", record={}):
        """insert dictionary to table"""
        if not (table_name and record):
            return None

        if not table_name.upper().startswith("LOG_"):
            record["zz_time"] = f"{self.cursor().now()}"
            record["zz_mode"] = "i"
            # check foreign keys
            foreign_keys_list = self.db_schema.get_primary_tables(table_name, record)
            for x in foreign_keys_list:
                if (
                    self.get(x["primary_table"], "%(primary_column)s='%(child_value)s'" % x, x["child_value"])
                    == {}
                ):
                    self.last_sql_error = (
                        "Foreign key error for insert:"
                        + f" For {table_name}"
                        + ".{child_column}".format(**x)
                        + " not found value '{child_value}' ".format(**x)
                        + "in table "
                        + x["primary_table"]
                        + ".{primary_column}".format(**x)
                    )
                    # print(self.last_sql_error)
                    self.last_error_data = x
                    return False

        table_columns = self.get_database_columns(table_name)
        primary_key_columns = self.get_primary_key_columns(table_name)
        columns_list = [x for x in record if x in table_columns]

        for x in primary_key_columns:
            if x not in columns_list:
                columns_list.append(x)
            primary_key_value = int_(record.get(x, 0))
            while (
                self.cursor(sql=f"select {x} from {table_name} where {x}={primary_key_value}").row_count() > 0
            ):
                primary_key_value += 1
            record[x] = primary_key_value

        sql = (
            f"insert into {table_name} ("
            + ",".join([x for x in columns_list])
            + ") values ("
            + ",".join(["%s" for x in columns_list])
            + ")"
        )

        if self.db_engine_name == "sqlite3":
            sql = self._sqlite_patch(sql, record, table_columns)

        self._check_record_for_numbers(table_name, record)
        data = [record[x] for x in columns_list]

        self._cursor(sql, data)

        if self.last_sql_error:
            return False
        else:
            if not table_name.upper().startswith("LOG_") and table_name.upper() != "PLATFORM":
                self.insert("log_" + table_name, record)
            return True

    def update(self, table_name="", record={}):
        """update from dictionary to table"""
        if not (table_name and record):
            return None

        table_columns = self.get_database_columns(table_name)
        primary_key_columns = self.get_primary_key_columns(table_name)

        if not table_name.upper().startswith("LOG_"):
            record["zz_time"] = f"{self.cursor().now()}"
            record["zz_mode"] = "u"
            foreign_keys_list = self.db_schema.get_primary_tables(table_name, record)
            for x in foreign_keys_list:
                if x["child_column"] not in record:  # column not going to change - skip checking
                    continue
                if self.get(x["primary_table"], "%(primary_column)s='%(child_value)s'" % x) == {}:
                    self.last_sql_error = (
                        "Foreign key error for update:"
                        + f" For {table_name}"
                        + ".{child_column}".format(**x)
                        + " not found value '{child_value}' ".format(**x)
                        + "in table "
                        + x["primary_table"]
                        + ".{primary_column}".format(**x)
                    )
                    self.last_error_data = x
                    return False

        columns_list = [x for x in record if x in table_columns]
        if is_sub_list(primary_key_columns.keys(), record.keys()):
            sql = f"update {table_name} set " + ",".join(
                [" %s=%%s " % x for x in record if x not in primary_key_columns and x in columns_list]
            )
            sql += " where " + " and ".join([f"{x}=%s " for x in primary_key_columns])
            if self.db_engine_name == "sqlite3":
                sql = self._sqlite_patch(sql, record, table_columns)

            self._check_record_for_numbers(table_name, record)
            data = [record[x] for x in record if x not in primary_key_columns and x in columns_list]
            data += [record[x] for x in primary_key_columns]

            self._cursor(sql, data)

            if self.last_sql_error:
                return False
            else:
                if not table_name.upper().startswith("LOG_") and table_name.upper() != "PLATFORM":
                    self.insert("log_" + table_name, record)
                return True

    def delete(self, table_name="", record={}):
        if not (table_name and record):
            return None
        self.last_error_data = {}
        for x in self.db_schema.get_child_tables(table_name, record):
            rez = self._cursor(
                "select 1 from {child_table} where {child_column}='{parent_value}'".format(**x)
            )
            if {} != rez:

                self.last_sql_error = (
                    "Foreign key error for delete:"
                    + f" Value in {table_name}"
                    + ".{parent_column}".format(**x)
                    + "={parent_value}".format(**x)
                    + " can not to be deleted, beacause "
                    + " it used in table {child_table}.{child_column}".format(**x)
                )
                self.last_error_data = x

                return False
        table_columns = self.get_database_columns(table_name)
        columns_list = [x for x in record if x in table_columns]
        where_clause = " and ".join([f"{x}=%s " for x in columns_list])
        data = [record[x] for x in columns_list]

        select_sql = f"select * from {table_name} where {where_clause}"
        if self.db_engine_name == "sqlite3":
            select_sql = self._sqlite_patch(select_sql, record, table_columns)

        row_to_be_deleted = self._cursor(select_sql, data)
        for x in row_to_be_deleted:
            row_to_be_deleted[x]["zz_mode"] = "d"
            row_to_be_deleted[x]["zz_time"] = f"{self.cursor().now()}"
            self.insert("log_" + table_name, row_to_be_deleted[x])

        sql = f"delete from {table_name} where {where_clause}"
        if self.db_engine_name == "sqlite3":
            sql = self._sqlite_patch(sql, record, table_columns)

        self._cursor(sql, data)

        if self.last_sql_error:
            return False
        else:
            return True

    def get(self, table_name="", where="", column_name=""):
        """
        return value of given column or record dictionary
        from first row  given table_name for where condition
        """
        column_name = f"({column_name}) as ret " if column_name else "*"
        row = self._cursor(f"""select {column_name} from {table_name} where {where}""")
        if self.last_sql_error:
            return ""
        else:
            if row:
                if column_name == "*":
                    return row[0]
                else:
                    return row[0]["ret"]
        return {}

    def _dict_factory(self, cursor, row, sql):
        return {
            col[0]: f"{row[idx] if row[idx] is not None else ''}".rstrip()
            for idx, col in enumerate(cursor.description)
        }

    def _cursor(self, sql, data=[]):
        self.last_sql_error = ""
        self.last_sql = ""
        self.last_record = ""
        _rows = {}
        try:
            _work_cursor = self.connection.cursor()
            if data:
                _work_cursor.execute(sql, data)
            else:
                _work_cursor.execute(sql)
            if _work_cursor.description:
                i = 0
                for x in _work_cursor.fetchall():
                    _rows[i] = self._dict_factory(_work_cursor, x, sql)
                    i += 1
        except self.db_api_engine.Error as err:
            self.last_sql_error = {err}
            self.last_sql = sql
            self.last_record = "!".join([f"{x}" for x in data])
            _rows = {0: {}}
        return _rows

    def cursor(self, sql="", table_name="", order="", where="", data=[], cache_flag=True):
        # TODO - sanitaze where and order
        return self.db_cursor_class(
            self,
            sql,
            table_name=table_name,
            order=order,
            where=where,
            data=data,
            cache_flag=cache_flag,
        )



#    Copyright (C) 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""

"""
if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")

    from tests import test_schema

    test_schema.test_schema()

    from demo import demo_01

    demo_01.demo()


class ZzDbSchema:
    """
    {"tables": {"table_name": {"columns": {}, "indexes": {...}}}}
    """

    def __init__(self, schema={}):
        self.schema = {"tables": {}}

    def add(
        self,
        table="",
        column="",
        datatype=None,
        datalen=None,
        datadec=None,
        to_table=None,
        to_column=None,
        related=None,
        pk=None,
        ai=None,
        uk=None,
    ):
        """
        table - database table name

        column - column name
        datatype - type
        datalen - lenght
        datadec - decimal precison

        to_table - foreign key table
        to_column -  foreign key column
        related - ?foreign column to show

        pk - primary key
        ai - autoincrement
        uk - unique
        """
        if not (table or column):
            return
        if table not in self.schema["tables"]:
            self.schema["tables"][table] = {"columns": {}, "indexes": {}}
        self.schema["tables"][table]["columns"][column] = {}

        self.schema["tables"][table]["columns"][column]["datatype"] = datatype
        self.schema["tables"][table]["columns"][column]["datalen"] = datalen
        self.schema["tables"][table]["columns"][column]["datadec"] = datadec
        self.schema["tables"][table]["columns"][column]["to_table"] = to_table
        self.schema["tables"][table]["columns"][column]["to_column"] = to_column
        self.schema["tables"][table]["columns"][column]["related"] = related
        self.schema["tables"][table]["columns"][column]["pk"] = pk
        self.schema["tables"][table]["columns"][column]["ai"] = ai
        self.schema["tables"][table]["columns"][column]["uk"] = uk

    def get_schema_table_attr(self, table="", column="", attr=""):
        """
        returs schema data for given table, column, attribute
            get_schema_table_attr(table_name) - all columns
            get_schema_table_attr(table_name,column_name) - given column
            get_schema_table_attr(table_name,column_name,"datalen") - given attribute
        """
        rez = self.schema.get("tables", {})
        if table == "":
            return rez
        rez = rez.get(table, {}).get("columns", {})
        if column == "":
            return rez
        rez = rez.get(column, {})
        if attr == "":
            return rez
        return rez.get(attr, "")

    def get_primary_tables(self, child_table, child_record):
        """
        returns list of foreign key tables and columns
        for given 'child_table' and 'child_record'

        used by ZzDb for integrity checking when INSERT/UPDATE
        """
        rez = []
        for child_column_name in self.get_schema_table_attr(child_table):
            child_column = self.get_schema_table_attr(child_table, child_column_name)
            if child_column.get("to_table") and child_column.get("to_column"):
                rez.append(
                    {
                        "primary_table": child_column.get("to_table"),
                        "primary_column": child_column.get("to_column"),
                        "child_column": child_column_name,
                        "child_value": child_record.get(child_column_name, ""),
                    }
                )
        return rez

    def get_child_tables(self, primary_table, primary_record):
        """
        returns list of foreign key tables and columns
        for given 'primary_table' and 'primary_record'

        used by ZzDb for integrity checking when DELETE
        """
        rez = []
        for linked_table_name in self.get_schema_table_attr():
            for linked_column_name in self.get_schema_table_attr(linked_table_name):
                linked_column = self.get_schema_table_attr(linked_table_name, linked_column_name)
                if linked_column.get("to_table") == primary_table and linked_column.get("to_column"):
                    parentCol = linked_column.get("to_column")
                    rez.append(
                        {
                            "child_table": linked_table_name,
                            "child_column": linked_column_name,
                            "parent_column": parentCol,
                            "parent_value": primary_record.get(parentCol),
                        }
                    )
        return rez



if __name__ == "__main__":
    import sys

    if "." not in sys.path:
        sys.path.insert(0, ".")

    # from demo.demo_mysql import demo
    from demo.demo_sqlite import demo

    demo()

import uuid
import re

from datetime import datetime

from zzdb.utils import int_


class Record:
    def __init__(self, t):
        self.t = t

    def __getattr__(self, name):
        return self.t.record(self.t.current_row()).get(name, f"column name '{name}' not found")


class ZzCursor:
    def __init__(self, zz_db, sql, table_name="", order="", where="", data=[], cache_flag=False):
        self.zz_db = zz_db
        self.tempName = "_%s" % uuid.uuid4().hex.replace("-", "")
        self.sql = sql
        self.table_name = table_name
        self.order = order
        self.primary_key_columns = []
        self.where = where
        self.data = data
        self.cache_flag = cache_flag
        self._rows = {}
        self._columns = []
        self._row_count = 0
        self._currentRow = 0
        self.refresh()
        self.r = Record(self)

    def get_table_names_sql(table_select_clause="", database_name=""):
        """returns sql statement (depending database) for select a list of all tables or given table"""
        pass

    def get_table_columns_sql(table_name="", where_clause="", database_name=""):
        """return database depending sql statement for select a list of all tables or given table"""
        pass

    def now(self):
        return datetime.now().strftime("%Y%m%d%H%M%S")

    def set_order(self, sort=""):
        """set order when cursor base on table"""
        self.order = sort
        return self

    def set_where(self, where=""):
        """set where condition when cursor base on table"""
        self.where = where
        return self

    def last_sql_error(self):
        return self.zz_db.last_sql_error

    def last_sql(self):
        return self.zz_db.last_sql

    def last_record(self):
        return self.zz_db.last_record

    def lastDeleteData(self):
        return self.zz_db.lastDeleteData

    def subFilter(self, column, text):
        if self.table_name:
            f = "".join(self.primary_key_columns)
            f += f",{column}" if column not in self.primary_key_columns else ""
            sql = f"select {f} from {self.table_name} where "
            if self.where:
                sql += f" {self.where} and "
            sql += f" ({self.prepColumnSearch(column,text)}) "
            if self.order:
                sql += f" order by {self.order}"
            error, _rows = self.zz_db._cursor(f"""{sql}""")
            return _rows
        return {}

    def prepColumnSearch(self, column, searchText="", placeHolder="before"):
        rez = []
        _or = []

        def _orAppend(rez):
            _or.append(f"({' and '.join(rez)})") if rez else ""

        mode = "like"
        for x in re.split(r"(\+|\-|\*)", searchText):
            if x == "+":
                mode = "like"
            elif x == "-":
                mode = "not like"
            elif x == "*":
                _orAppend(rez)
                rez = []
                mode = "like"
            elif x:
                # x=zzQ(x)
                if placeHolder == "before":
                    rez.append(f" {column} {mode} '%{x}%' ")
                else:
                    rez.append(f" '%{x}%' {mode} {column} ")
        _orAppend(rez)
        return " or ".join(_or)

    def update(self, data):
        if self.table_name:
            rez = self.zz_db.update(self.table_name, data)
            return rez

    def insert(self, data, refresh=True, where=True):
        if self.table_name:
            if self.where and where:
                for x in re.split(" and ", self.where, flags=re.IGNORECASE):
                    if x.count("=") == 1 and " OR " not in x.upper():
                        data[x.split("=")[0].strip()] = eval(x.split("=")[1])
            rez = self.zz_db.insert(self.table_name, data)
            if refresh and rez:
                self.refresh()
            return rez

    def delete(self, data):
        if self.table_name:
            rez = self.zz_db.delete(self.table_name, data)
            self.refresh()
            return rez

    def get(self, where="", column_name=""):
        if self.table_name:
            return self.zz_db.get(self.table_name, where, column_name)
        return {}

    def current_row(self):
        return self._currentRow

    def first(self):
        self._currentRow = 0

    def last(self):
        self.set_current_row(self.row_count() - 1)

    def next(self):
        self.set_current_row(self.current_row() + 1)

    def prev(self):
        self.set_current_row(self.current_row() - 1)

    def set_current_row(self, current_row):
        if current_row >= 0:
            if current_row < self.row_count():
                self._currentRow = current_row
            else:
                self.last()
        else:
            self._currentRow = 0

    def eof(self):
        return self._currentRow == self._row_count - 1

    def bof(self):
        return self._currentRow == 0

    def getPkRow(self, dataDic):
        pkName = [x for x in self.primary_key_columns][0]
        pkValue = str(dataDic[pkName])
        for x in range(self.row_count()):
            if self.record(x)[pkName] == pkValue:
                return x

    def get_primary_key_columns(self):
        return self.primary_key_columns[:]

    def refresh(self):
        self._currentRow = 0
        return self

    def row_count(self):
        return self._row_count

    def records(self):
        for x in range(self._row_count):
            self._currentRow = x
            yield self.record(x)

    def record(self, rowNumber, columns=[]):
        if rowNumber in self._rows:
            if columns:
                return {x: self._rows[rowNumber][x] for x in self._rows[rowNumber] if x in columns}
            else:
                return self._rows[rowNumber]
        else:
            return {}

    def get_columns(self):
        return [x for x in self.record(0)]


class ZzSqliteCursor(ZzCursor):
    def refresh(self):
        super().refresh()
        if self.table_name:
            self.primary_key_columns = [x for x in self.zz_db.get_primary_key_columns(self.table_name)]
            self.sql = f"select * from {self.table_name}"
            if self.where:
                self.sql += f" where {self.where}"
            if self.order:
                self.sql += f" order by {self.order}"
        self._rows = self.zz_db._cursor(f"""{self.sql}""", self.data)
        if self.zz_db.last_sql_error or self._rows == {}:
            self._row_count = -1
        else:
            self._row_count = len(self._rows)
        return self

    @staticmethod
    def get_table_names_sql(table_select_clause="", database_name=""):
        return f"""SELECT distinct tbl_name as table_name
                FROM sqlite_master
                WHERE  type = 'table' {table_select_clause} """

    @staticmethod
    def get_table_columns_sql(table_name="", where_clause="", database_name=""):
        if where_clause:
            where_clause = f" where {where_clause}"
        return f"""select
                        name,
                        type as datatype,
                        `notnull` as nn,
                        `dflt_value` as `default`
                    from PRAGMA_table_info("{table_name}")
                    {where_clause}
                    """


class ZzMysqlCursor(ZzCursor):
    def __del__(self):
        try:
            self.removeTempTable()
        except Exception:
            pass

    def setCacheFlag(self, flag):
        if flag != self.cacheFlag:
            self._rows = {}
        self.cacheFlag = flag

    def record(self, rowNumber, columns={}):
        if rowNumber >= 0 and rowNumber < self._row_count:
            if rowNumber not in self._rows:
                if self.primary_key_columns:
                    row = self.zz_db._cursor(
                        f"""select {self.table_name}.*
                                            from {self.table_name},{self.tempName}
                                            where {self.tempName}._rn_={rowNumber} and
                                            {' and '.join([f'{self.table_name}.{x}={self.tempName}.{x}'
                                                                for x in self.primary_key_columns])}
                                        """
                    )
                else:
                    row = self.zz_db._cursor(
                        f"""select *
                            from {self.tempName}
                            where _rn_={rowNumber}"""
                    )
                if self.zz_db.last_sql_error:
                    row = {0: {}}
                if self.cache_flag:
                    self._rows[rowNumber] = row[0]
                # print (rowNumber,row)
                rez = row[0]
            else:
                rez = self._rows[rowNumber]
            # return rez
            if columns:
                return {x: rez[x] for x in rez if x in columns}
            else:
                return rez

        else:
            return {}

    def getPkRow(self, dataDic):
        if self.primary_key_columns:
            sql = f"""select _rn_
                    from {self.tempName}
                    where {' and '.join([f'{x}=%s' for x in self.primary_key_columns])}
                        """
            data = [dataDic[x] for x in self.primary_key_columns]
            error, row = self.zz_db._cursor(sql, data)
            if error:
                return -1
            else:
                if row:
                    return int_(row[0]["_rn_"])
                else:
                    return -1
        else:
            return -1

    def removeTempTable(self):
        if self.tempName and self.zz_db:
            self.zz_db._cursor(f"drop temporary table IF EXISTS {self.tempName} ")
        self._rows = {}
        self._row_count = -1

    def close(self):
        self.removeTempTable()
        self.zz_db = None

    def refresh(self):
        super().refresh()
        if self.table_name:
            self.primary_key_columns = [x for x in self.zz_db.get_primary_key_columns(self.table_name)]
            if self.primary_key_columns:
                self.sql = f"select {','.join(self.primary_key_columns)} from {self.table_name}"
            else:
                self.sql = f"select * from {self.table_name}"
            if self.where:
                self.sql += f" where {self.where}"
            if self.order:
                self.sql += f" order by {self.order}"
        self.removeTempTable()
        if self.sql.strip().upper().startswith("SELECT"):
            self.zz_db._cursor(
                f"""
                            create temporary table {self.tempName}
                            (
                                select @i := @i + 1 AS _rn_, z1.*
                                from  ({self.sql}) z1, (select @i:=-1) z2
                            )
                                """,
                self.data,
            )
        else:
            self.zz_db._cursor(f"""{self.sql}""", self.data)
        if self.zz_db.last_sql_error:
            self._row_count = -1
        else:
            self.zz_db._cursor(f"""alter table {self.tempName} add index (_rn_)""")
            _row = self.zz_db._cursor(f"select count(1) as rc from {self.tempName}")
            if self.zz_db.last_sql_error or _row == []:
                self._row_count = -1
            else:
                self._row_count = int_(_row[0]["rc"])
        return self

    @staticmethod
    def get_table_names_sql(table_select_clause="", database_name=""):
        return f"""select distinct table_name
                   FROM INFORMATION_SCHEMA.TABLES
                   WHERE table_schema='{database_name}' and
                       TABLE_TYPE<>'VIEW' {table_select_clause}
                """

    @staticmethod
    def get_table_columns_sql(table_name="", where_clause="", database_name=""):
        if where_clause:
            where_clause = f" and {where_clause}"
        return f"""select
                        column_name as name,
                        data_type as datatype,
                        column_type ,
                        case when character_maximum_length<>0
                                then character_maximum_length
                                else numeric_precision
                        end as datalen,
                        numeric_scale as datadec
                    FROM INFORMATION_SCHEMA.COLUMNS
                    WHERE table_name = '{table_name}' and
                    table_schema='{database_name}'
                    {where_clause}
                    """


class ZzPostgresqlCursor(ZzCursor):
    def refresh(self):
        super().refresh()
        if self.table_name:
            self.primary_key_columns = [x for x in self.zz_db.get_primary_key_columns(self.table_name)]
            self.sql = f"select * from {self.table_name}"
            if self.where:
                self.sql += f" where {self.where}"
            if self.order:
                self.sql += f" order by {self.order}"
        self._rows = self.zz_db._cursor(f"""{self.sql}""", self.data)
        if self.zz_db.last_sql_error or self._rows == {}:
            self._row_count = -1
        else:
            self._row_count = len(self._rows)
        return self

    @staticmethod
    def get_table_names_sql(table_select_clause="", database_name=""):
        return f"""SELECT table_name
                    FROM information_schema.columns
                    where table_catalog='{database_name}' and
                        table_schema='public' {"and" if table_select_clause else ""} {table_select_clause}
                     """

    @staticmethod
    def get_table_columns_sql(table_name="", where_clause="", database_name=""):
        if where_clause:
            where_clause = f" and {where_clause}"
        return f"""
                    select
                        isc.column_name as name,
                        isc.data_type as datatype,
                        case when isc.character_maximum_length<>0
                                then isc.character_maximum_length
                                else isc.numeric_precision
                        end as datalen,
                        isc.numeric_scale as datadec,
                        column_key

                    from information_schema.columns isc

                    left join (
                                select 'PRI' as column_key,
                                    is_ccu.column_name,
                                    is_ccu.table_name,
                                    is_ccu.constraint_catalog as table_catalog
                                from information_schema.constraint_column_usage is_ccu,
                                    information_schema.table_constraints  is_tc
                                where is_ccu.constraint_name=is_tc.constraint_name and
                                    is_ccu.constraint_catalog = is_tc.constraint_catalog and
                                    is_ccu.table_name = is_tc.table_name
                                ) ist
                    on
                        isc.table_catalog = ist.table_catalog and
                        isc.table_name = ist.table_name and
                        isc.column_name = ist.column_name

                    where
                        isc.table_catalog = '{database_name}' and
                        isc.table_schema = 'public' and
                        isc.table_name = '{table_name}'
                        {where_clause}

                    order by ordinal_position
                    """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zzdb']

package_data = \
{'': ['*']}

install_requires = \
['mysql-connector-python>=8.0.26,<9.0.0', 'psycopg2-binary>=2.9.1,<3.0.0']

setup_kwargs = {
    'name': 'zzdb',
    'version': '0.1.1',
    'description': 'python DB-API wrapper',
    'long_description': '# The light Python DB API wrapper with some ORM functions\n\n## Features:\n ---\n## Connect\n```python\nfrom zzdb.db import ZzDb\n\ndatabase_sqlite = ZzDb("sqlite3", database_name=":memory:")\n\ndatabase_mysql = ZzDb(\n    "mysql",\n    database_name="zztest",\n    host="0.0.0.0",\n    port="3308",\n    user="root",\n    password="zztest"\n)\n\ndatabase_postgresql = ZzDb(\n    "postgresql",\n    database_name="zztest1",\n    host="0.0.0.0",\n    port=5432,\n    user="zzuser",\n    password="zztest"\n)\n```\n---\n## Define & migrate database schema (ADD COLUMN only).\n```python\nzzdb.schema import ZzDbSchema\n\nschema = ZzDbSchema()\n\nschema.add(table="topic_table", column="uid", datatype="int", datalen=9, pk=True)\nschema.add(table="topic_table", column="name", datatype="varchar", datalen=100)\n\nschema.add(table="message_table", column="uid", datatype="int", datalen=9, pk=True)\nschema.add(table="message_table", column="message", datatype="varchar", datalen=100)\nschema.add(\n    table="message_table",\n    column="parent_uid",\n    to_table="topic_table",\n    to_column="uid",\n    related="name"\n)\n\ndatabase.set_schema(schema)\n```\n---\n## INSERT, UPDATE, DELETE\n```python\ndatabase.insert("topic_table", {"name": "topic 0"})\ndatabase.insert("topic_table", {"name": "topic 1"})\ndatabase.insert("topic_table", {"name": "topic 2"})\ndatabase.insert("topic_table", {"name": "topic 3"})\n\ndatabase.insert("message_table", {"message": "Message 0 in 0", "parent_uid": 0})\ndatabase.insert("message_table", {"message": "Message 1 in 0", "parent_uid": 0})\ndatabase.insert("message_table", {"message": "Message 0 in 1", "parent_uid": 1})\ndatabase.insert("message_table", {"message": "Message 1 in 1", "parent_uid": 1})\n\n# this returns False because there is no value 2 in topic_table.id - schema works!\ndatabase.insert("message_table", {"message": "Message 1 in 1", "parent_uid": 2})\n\n\ndatabase.delete("message_table", {"uid": 2})\n\ndatabase.update("message_table", {"uid": 0, "message": "updated message"})\n```\n---\n## Cursor\n```python\ncursor = database.cursor(table_name="topic_table")\ncursor = database.cursor(\n    table_name="topic_table",\n    where=" name like \'%2%\'",\n    order="name desc"\n)\ncursor.insert({"name": "insert record via cursor"})\ncursor.delete({"uid": 2})\ncursor.update({"uid": 0, "message": "updated message"})\n\ncursor = database.cursor(sql="select name from topic_table")\n\nfor x in cursor.records():\n    print(x)\n\ncursor.record(0)[\'name\']\ncursor.row_count()\ncursor.first()\ncursor.last()\ncursor.next()\ncursor.prev()\ncursor.bof()\ncursor.eof()\n```',
    'author': 'Andrei Puchko',
    'author_email': 'andrei.puchko@gmx.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
